#include <iostream>

#include "Edge.h"
#include "Vertex.h"
#include "Face.h"

#include <maya/MVector.h>
#include <maya/MMatrix.h>
#include <maya/MPoint.h>

#include <vector>
#include <float.h>
#include <assert.h>
#include <algorithm>

using namespace std;

Edge::Edge(Vertex *_a, Vertex *_b) : 
 v0(_a), 
 v1(_b),
 f0(NULL),
 f1(NULL),
 normal(MVector()),
 cost(DBL_MAX),
 valid(false),
 removed(false) {
   v0->edges.push_back(this);
   v1->edges.push_back(this);
}

Edge::Edge(const Edge &rhs) : 
 v0(rhs.v0), 
 v1(rhs.v1),
 f0(rhs.f0),
 f1(rhs.f1),
 cost(rhs.cost),
 newpos(rhs.newpos),
 normal(rhs.normal),
 Q(rhs.Q),
 valid(rhs.valid),
 removed(rhs.removed) {

}

const Edge &Edge::operator=(const Edge &rhs) {
   v0 = rhs.v0;
   v1 = rhs.v1;
   f0 = rhs.f0;
   f1 = rhs.f1;
   cost = rhs.cost;
   normal = rhs.normal;
   newpos = rhs.newpos;
   Q = rhs.Q;
   valid = rhs.valid;
   removed = rhs.removed;
   return *this;
}

void Edge::ComputeCost() {
   if (removed)
      return;
   enum {BEST, LINE, LOCAL} policy = BEST;

   if (!v0 || !v1) {
      cost = DBL_MAX;
      return;
   }


   Q = v0->Q + v1->Q;
   normal = v0->normal + v1->normal;
   normal.normalize();
   
   // TODO: normalize normal

   if ((!f0 || !f1) || f0->normal * f1->normal == 0)
      policy = LINE;

   switch (policy) {
   case BEST:
      if (findBestFit())
         break;
   case LINE:
      if (findLineFit())
         break;
   case LOCAL:
      if (!findLocalFit())
         cerr << "ERROR: Failed to find local fit in Edge::ComputeCost" << endl;

      break;
   default:
      assert(0);
   }
   
   double x = newpos.x;
   double y = newpos.y;
   double z = newpos.z;

   // this = v^T Q v (as used by Garland)
   cost = x*x*Q(0,0) + 2*x*y*Q(0,1) + 2*x*z*Q(0,2) + 2*x*Q(0,3)
      +   y*y*Q(1,1) + 2*y*z*Q(1,2) +   2*y*Q(1,3)
      +   z*z*Q(2,2) +   2*z*Q(2,3)
      +       Q(3,3);

   // TODO: Heap Stuff
}

bool Edge::findBestFit() {
    // this comes from Garland
   MMatrix K = Q;
   K(3,0) = K(3,1) = K(3,2) = 0.0;  K(3,3) = 1;
   if (K.isSingular())
      return false;

   MMatrix M = K.inverse(); 

   newpos = MPoint(M(0,3), M(1,3), M(2,3));
   return true;
}
bool Edge::findLineFit() {
   MPoint p0 = v0->pos;
   MPoint p1 = v1->pos;

   MVector d = p0 - p1;
   MVector Qv2 = Q * p1;
   MVector Qd = Q * d;

   double denom = 2 * d * Qd;

   if (denom == 0.0)
      return false;

   double a = (d * Qv2 + MVector(p1) * Qd) / denom;

   if (a < 0.0)
      a = 0.0;
   if (a > 1.0)
      a = 1.0;

   newpos = a * d + p1;
   return true;
}

bool Edge::findLocalFit() {
   newpos = (v0->pos + v1->pos) / 2.0;
   return true;
}

int Edge::Collapse() {
   if (removed) {
      cerr << "Trying to collapse a removed edge" << endl;
      return -1;   
   }
   removed = true;
   int numRemoved = 0;
   int i, fNdx;

   Vertex *lives, *dies;
   Face *left, *right, *kill;

   Edge *l2k, *r2k;

   lives = v0;
   dies = v1;

   for (fNdx = 0; fNdx < 2; fNdx++) {
      kill = faces[fNdx];
      if (kill) {
         kill->removed = true;
         Edge **faceEdges = kill->edges;

         for (i = 0; i < 3; i++)
            if (faceEdges[i] == this)
               break;

         if (faceEdges[(i + 1) % 3]->Contains(lives)) {
            l2k = faceEdges[(i + 1) % 3];
            r2k = faceEdges[(i + 2) % 3];
         } else {
            l2k = faceEdges[(i + 2) % 3];
            r2k = faceEdges[(i + 1) % 3];
         }

         left = l2k->GetOtherFace(kill);
         right = r2k->GetOtherFace(kill);
         r2k->removed = true;

         if (!left && !right)
            l2k->removed = true;
         else {
            l2k->RemapFace(kill, right);

            if (right) {
               right->RemapEdge(r2k, l2k);
            }
         }
         l2k->Verify();
         numRemoved++;
      }
   }
   dies->RemapTo(lives); 
   /*
   remove(lives->edges.begin(), lives->edges.end(), this);
   remove(lives->faces.begin(), lives->faces.end(), f0);
   remove(lives->faces.begin(), lives->faces.end(), f1);
   */
   lives->pos = newpos;
   return numRemoved;
}

bool Edge::Valid() {
   if (removed)
      return false;

   // TODO: Expand this to original code

}

Face *Edge::GetOtherFace(const Face* const f) const {
   if (removed)
      cerr << "Trying to getOtherFace accross a removed edge!" << endl;

   if (f0 && f0 == f)
      return f1;
   else if (f1 && f1 == f)
      return f0;
   cerr << "WARNING: Got to the bottom of Edge::GetOtherFace" << endl;
   return 0;
}

void Edge::RemapTo(Edge *eNew, Face *fOld) {
   if (eNew->removed) {
      cerr << "ERROR: Remapping Edge to a removed edge!" << endl;
      return;
   } else if (eNew == this)
      cerr << "ERROR: Trying to remap edge to itself" << endl;
   else {
      removed = true;
      Face *left = GetOtherFace(fOld);
      Face *right = eNew->GetOtherFace(fOld);

      if (!left && !right)
         eNew->removed = true;
      else {
         eNew->RemapFace(fOld, left);
         if (left)
            left->RemapEdge(this, eNew);
      }
      eNew->Verify();
   }
}


void Edge::Verify() {
   if (removed)
      return;
   if (f0) {
      if (!f0->ContainsEdge(this))
         cerr << "Edge has face, but face does not have edge!" << endl;
      else
         f0->Verify();
   }

   if (f1) {
      if (!f1->ContainsEdge(this))
         cerr << "Edge has face, but face does not have edge!" << endl;
      else
         f1->Verify();
   }
}

void Edge::AddFace(Face *f) {
   if (f0 == NULL)
      f0 = f;
   else if (f1 == NULL)
      f1 = f;
   else
      cerr << "Trying to add face to a full edge..." << endl;
}

bool Edge::Contains(Vertex *v) {
   return v0 == v || v1 == v;
}

void Edge::RemapFace(Face *fOld, Face *fNew) {
   if (f0 == fOld)
      f0 = fNew;
   else if (f1 == fOld)
      f1 = fNew;
   else
      cerr << "Trying to remap a face which isn't on this edge..." << endl;
}

void Edge::RemapVertex(Vertex *vOld, Vertex *vNew) {
   if (v0 == vOld) {
      v0 = vNew;
      vNew->edges.push_back(this);
   } else if (v1 == vOld) {
      v1 = vNew;
      vNew->edges.push_back(this);
   } else
      cerr << "Trying to remap vertex which isn't on this edge..." << endl;
}