#include <iostream>
#include <algorithm>
#include <maya/MGlobal.h>
#include <maya/MPoint.h>
#include <maya/MVector.h>
#include <maya/MString.h>
#include <maya/MDagPath.h>
#include <maya/MPointArray.h>
#include <maya/MFloatPointArray.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItMeshEdge.h>
#include <maya/MItMeshVertex.h>
#include <maya/MItMeshFaceVertex.h>
#include <maya/MFnMesh.h>
#include <maya/MMessage.h>
#include "QEMMesh.h"
#include "util.h"
#include "Edge.h"

using namespace std;

QEMMesh::QEMMesh() {

}

void QEMMesh::AddNode(MDagPath &node) {
   vector<Vertex *> tmp_verts; // Tmp list of verts for just this node
   vector<Face *> tmp_faces;
   vector<Edge *> tmp_edges;

   // Create all of the vertices for the node
   MFnMesh meshFn(node);
   MPointArray points;
   MVector normal;
   MFloatVectorArray normals;

   meshFn.getPoints(points);
   meshFn.getNormals(normals);
   for (uint i = 0; i < points.length(); i++) {
      tmp_verts.push_back(new Vertex(points[i]));
      tmp_verts.back()->normal = MVector(normals[i]);
   }

   for (int i = 0; i < meshFn.numEdges(); i++) {
      // Build edges
      int2 vertIds;
      meshFn.getEdgeVertices(i, vertIds);
      Edge *e = new Edge(tmp_verts[vertIds[0]], tmp_verts[vertIds[1]]);
      tmp_edges.push_back(e);
   }

   // For now, just grab one triangle from each polygon
   MItMeshPolygon pItr(node, MObject::kNullObj);

   for (; !pItr.isDone(); pItr.next()) {
      // Build a face from 3 verts - Assumes pretriangulation
      MIntArray edgeList;
      Face *f = new Face(tmp_verts[pItr.vertexIndex(0)], tmp_verts[pItr.vertexIndex(1)], tmp_verts[pItr.vertexIndex(2)]);
      pItr.getNormal(normal);
      f->normal = normal;
      tmp_faces.push_back(f);

      pItr.getEdges(edgeList);
      f->x = tmp_edges[edgeList[0]];
      f->x->AddFace(f);
      f->y = tmp_edges[edgeList[1]];
      f->y->AddFace(f);
      f->z = tmp_edges[edgeList[2]];
      f->z->AddFace(f);

      f->ComputeCost();
   }

   // Now that we've built everything, add them back to the main lists
   verts.insert(verts.end(), tmp_verts.begin(), tmp_verts.end());
   edges.insert(edges.end(), tmp_edges.begin(), tmp_edges.end());
   faces.insert(faces.end(), tmp_faces.begin(), tmp_faces.end());

   // Additonally, push the edges onto the heap
   vector<Edge *>::iterator itr;
   for (itr = tmp_edges.begin(); itr != tmp_edges.end(); itr++) {
      (*itr)->ComputeCost();
      heap.push_back(*itr);
      push_heap(heap.begin(), heap.end(), edge_ptr_cmp());
   }
}

bool QEMMesh::CollapseTopEdge() {
   if (heap.empty())
      return false;
   Edge *e = heap[0];
   pop_heap(heap.begin(), heap.end(), edge_ptr_cmp());
   heap.pop_back();

   cout << "Collapsing edge with endpoints " << e->v0->pos << " " << e->v1->pos << endl;
   int ret = e->Collapse();
   cout << "Collapsed " << ret << " items" << endl;

   // quick and dirty check to remove removed items from the heap
   vector<Edge *>::iterator itr;
   for (itr = heap.begin(); itr != heap.end(); itr++) {
      if ((*itr)->removed) {
         heap.erase(itr);
         // We should actually delete *itr too.
         itr = heap.begin();
      }
   }

   make_heap(heap.begin(), heap.end(), edge_ptr_cmp());
   delete e;
   return true;
}

void QEMMesh::ReplaceMesh(MDagPath &node) {
   MStatus stat;
   node.extendToShape();
   MFnMesh meshFn(node, &stat);

   if (!stat)
      cerr << stat << endl;
   MFloatPointArray points;
   MIntArray polyCounts;
   MIntArray polyConnects;

   int vertNum = 0;
   int polyNum = 0;
   
   // Construct mapping of pointer address to array location, to be used as ndx
   map<Vertex *, uint> mapping;
   uint id = 0;

   vector<Face *>::iterator fItr;
   for (fItr = faces.begin(); fItr != faces.end(); fItr++) {
      if (!(*fItr)->removed) {
         polyNum++;
         polyCounts.append(3);

         pair<map<Vertex *, uint>::iterator,bool> ret;
         ret = mapping.insert(pair<Vertex*, uint>((*fItr)->a, 0));
         if (ret.second)
            mapping[(*fItr)->a] = id++;

         ret = mapping.insert(pair<Vertex*, uint>((*fItr)->b, 0));
         if (ret.second)
            mapping[(*fItr)->b] = id++;

         ret = mapping.insert(pair<Vertex*, uint>((*fItr)->c, 0));
         if (ret.second)
            mapping[(*fItr)->c] = id++;

         polyConnects.append(mapping[(*fItr)->a]);
         polyConnects.append(mapping[(*fItr)->b]);
         polyConnects.append(mapping[(*fItr)->c]);
      }
   }

   points.setLength(mapping.size());
   map<Vertex *, uint>::iterator mItr;
   for (mItr = mapping.begin(); mItr != mapping.end(); mItr++) {
      float xyzw[4];
      mItr->first->pos.get(xyzw);
      points[mItr->second] = MFloatPoint(xyzw);
   }
   vertNum = points.length();

   cout << "DEBUG: polyNum is " << polyNum << endl;
   cout << "DEBUG: countsLen is " << polyCounts.length() << endl;
   cout << "DEBUG: connectsLen is " << polyConnects.length() << endl;

   stat = meshFn.createInPlace(vertNum, polyNum, points, polyCounts, polyConnects);
   if (!stat) {
      cerr << "Error creating replacement mesh" << endl;
      cerr << stat << endl;
   }
   meshFn.syncObject();
   meshFn.updateSurface();
}